/*
    pinger -- GTK+/ncurses multi-ping utility
    Copyright (C) 2002 Petr Koloros

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "parse.h"

/* Parse one line of config file ("name = value" format)
 * and store values into "name" and "value" */
void parse_line(char *line, char *name, char *value) {
  char *value_part;
  int value_part_index;

  value_part = strchr(line, (int)'=');
  if (value_part == NULL) {
    qprint("%s %s\n", _("Error: incomplete line:"), line);
    exit(1);
  }
  strcpy(value, 1 + value_part);
  value_part_index = strlen(line) - strlen(value) - 1;
  if (value_part_index >= 1) {
    strncpy(name, line, (size_t)value_part_index);
    name[value_part_index] = '\0';
  } else value[0] = '\0';
  
  if (value[strlen(value) - 1] == '\n') value[strlen(value) - 1] = '\0';
}

/* Check if byte (in string form) is real byte value (0-255) */
int check_byte(char *byte) {
  int byte_val;
  
  byte_val = atoi(byte);
  if ((byte_val < 0) || (byte_val > 255))  return 1;
  else return 0;
}

/* Check if valid option value */
int check_opt_value(char * value) {
  int intval;
  
  intval = atoi(value);
  if ((intval < 0) || (intval > MAX_OPT_VALUE)) return 1;
  else return 0;

}
  
/* Check if IP ("XXX.XXX.XXX.XXX" in string form) is real IPv4 IP address */
int valid_ip(char *ip) {

  int len, exit_now, cnt, dot_cnt, byte_cnt;
  char byte[BYTE_STR_LEN];  
  
  byte[0] = '\0';
  if ((len = strlen(ip)) <= 15) {
    for (cnt = 0, dot_cnt = 0, byte_cnt = 0,
        exit_now = 0; cnt < len; cnt++) {
      if (ip[cnt] == '.') {
        if (byte_cnt == 0) exit_now = 1;
        else {
          byte[byte_cnt] = '\0';
          if (check_byte(byte)) exit_now = 1;
          byte_cnt = 0;
        }
        dot_cnt++;
      } 
      else if (((int)ip[cnt] < (int)'0') || 
          ((int)ip[cnt] > (int)'9')) exit_now = 1;
      else {
        byte[byte_cnt] = ip[cnt];
        if (++byte_cnt == 4) exit_now = 1;
      }
      if (exit_now) cnt = len;
    }
    if (dot_cnt != 3) exit_now = 1;
  } else exit_now = 1;
  /* check last part of IP */
  byte[byte_cnt] = '\0';
  if (check_byte(byte)) exit_now = 1;
  /* if exit_now evaluated, exit ! */
  if (exit_now) return 1;
  else return 0;
}

/* parse value, which should be in this format:
 * hostname,pause_between_OK_pings,pause_between_NOREPLY_pings,pause_of_long_delay
 * It work also with only one (first) parameter. */
int parse_value (char *value, char *name, int *ok_delay, int *noreply_delay,
        int *long_delay) {
  char *tmp;

  tmp = strtok(value, ",");
  if (tmp == NULL) {
    name = NULL;
    return 1;
  } else strcpy(name, tmp);

  tmp = strtok(NULL, ",");
  if (tmp != NULL) { 
    *ok_delay = atoi(tmp);
    if (*ok_delay < MIN_DELAY) *ok_delay = MIN_DELAY;
    if (*ok_delay > MAX_DELAY) *ok_delay = MAX_DELAY;
  }
  else *ok_delay = delay_between_pings_ms;
  
  tmp = strtok(NULL, ",");
  if (tmp != NULL) {
    *noreply_delay = atoi(tmp);
    if (*noreply_delay < MIN_DELAY) *noreply_delay = MIN_DELAY;
    if (*noreply_delay > MAX_DELAY) *noreply_delay = MAX_DELAY;
  }
  else *noreply_delay = host_timeout_ms;

  tmp = strtok(NULL, ",");
  if (tmp != NULL) {
    *long_delay = atoi(tmp);
    if (*long_delay < MIN_DELAY) *long_delay = MIN_DELAY;
    if (*long_delay > MAX_DELAY) *long_delay = MAX_DELAY;
  }
  else *long_delay = MIN_DELAY;

  return 0;
}

/* Read config file and fill up host structure with these data.
 * Also open raw sockets for the hosts. */
void read_config_file (hosts_data *hosts, filelist *cfg_filenames) {

  char *homestr;  
  char cfg_path[512];
  FILE *cfg_file;
  char line[MAX_CFGLINE_LEN];
  char name[MAX_CFGLINE_LEN];
  char value[MAX_CFGLINE_LEN];
  char ipstr[MAX_CFGLINE_LEN];
  char val_name[MAX_CFGLINE_LEN];
  int ok_delay;
  int noreply_delay;
  int long_delay;
  int host_max, host_start, host_idx;
  int section;
  int int_value;
  int end_parsing = 0;
  filelist *cfgfilesidx;
  int cfg_file_opened = 0;
  host_data *host_info;
  titlelist *tlist;
  int tlistcnt = 0;
  int std_title_cnt = 1;
  int title_set;
  int set_nr = 0;
  int id_in_set = 0;
  int host_ininitalized_curr_file = 0;
  int stdcntlen;
  char *recent_title = NULL;
  char tmpstr[MIN_TITLE_LEN]; /* temporary string for 2 digits, used for std title */

  enum sections {
    INIT,
    OPTIONS,
    HOSTS
  };

  host_max = hosts->host_max;
  host_start = host_max;
  cfgfilesidx = cfg_filenames;
  host_info = hosts->host_info;
    
  homestr = getenv("HOME");
  if (!homestr) {
    qprint("%s 'HOME'.\n", _("Cannot get environment variable"));
    exit(1);
  }

  strcpy(cfg_path, homestr);
  strcat(cfg_path, "/");
  if (cfg_filenames == NULL)
    strcat(cfg_path, CFG_FILENAME);
  else
    strncpy(cfg_path, cfgfilesidx->filename, sizeof(cfg_path) - 1);
  while (!end_parsing) {
    qprint("%s %s\n", _("Reading configuration file"), cfg_path);
    host_ininitalized_curr_file = 0;
    cfg_file_opened = 0;
    cfg_file = fopen(cfg_path, "r");
    if (!cfg_file) {
      qprint("%s (%s)!\n", _("Cannot open config file"), cfg_path);
      if (cfgfilesidx != NULL) {
        cfgfilesidx = cfgfilesidx->next;
        if (cfgfilesidx == NULL) end_parsing =1;
        else strncpy(cfg_path, cfgfilesidx->filename, sizeof(cfg_path) - 1);
        break;
      }
      qprint("--\n%s\n", _("Creating new one in your home directory ..."));
      cfg_file = fopen(cfg_path, "w");
      if (!cfg_file) {
        qprint("%s. %s\n", _("Cannot create file"), _("Please check your create privileges for this directory!"));
        exit(1);
      } else {
        fprintf(cfg_file, "# %s\n", _("Pinger configuration file"));
        fprintf(cfg_file, "# <IP>=<%s>[,ok_delay,noreply_delay,long_delay]\n", _("Description"));
        fprintf(cfg_file, "#\n# ok_delay - %s\n", _("when ping successful, wait this time before sending next ping request."));
        fprintf(cfg_file, "# noreply_delay - %s\n", _("when ping failed, wait this time before sending next ping request."));
        fprintf(cfg_file, "# long_delay - %s\n", _("delay of reply after which is host marked as delayed."));
        fprintf(cfg_file, "\nOPTIONS SECTION:\n");
        fprintf(cfg_file, "\n# %s\n", _("display refresh interval"));
        fprintf(cfg_file, "#refresh_interval_ms=1000\n");
        fprintf(cfg_file, "\n# %s\n", _("successful pings' next request interval"));
        fprintf(cfg_file, "#delay_between_pings_ms=3000\n");
        fprintf(cfg_file, "\n# %s\n", _("failed pings' next request interval"));
        fprintf(cfg_file, "#host_timeout_ms=10000\n");
        fprintf(cfg_file, "\n# %s\n", _("delay of reply after which is host marked as delayed, if not specified in IP line"));
        fprintf(cfg_file, "#long_delay_ms=5000\n");
        fprintf(cfg_file, "\n# %s\n", _("title of this set of hosts. Character %% is not allowed here!"));
        fprintf(cfg_file, "title=standard set\n");
        fprintf(cfg_file, "\nHOSTS SECTION:\n");
        fprintf(cfg_file, "212.80.76.18=seznam\n");
        fprintf(cfg_file, "216.239.57.100=google,3000,10000,2000\n");
        fclose(cfg_file);
        if (chown(cfg_path, getuid(), getgid()))
          qprint("%s %s\n", _("Cannot set configuration file owner to current user:"), strerror(errno));
        cfg_file = fopen(cfg_path, "r");
        if (!cfg_file) {
          qprint("%s\n", _("Cannot reopen just created config file!"));
          exit(1);
        }
        qprint("%s (%s)\n", _("Config file created successfuly"),cfg_path);
      }
    } else qprint("%s %s\n", _("Using file"), cfg_path);
    cfg_file_opened = 1; 

    /* read from cfg file */
    title_set = 0;
    while ((fgets(line, MAX_CFGLINE_LEN, cfg_file) != NULL ) && (host_max < MAX_HOSTS)) {
      if ((!(line[0] == '\n')) && (!(line[0] == '#'))) {
        /* check SECTION */
        if (strcmp("OPTIONS SECTION:\n", line) == 0) {
          qprint("### %s\n", _("found options section"));
          section = OPTIONS;
        } else if (strcmp("HOSTS SECTION:\n", line) == 0) {
          qprint("### %s\n", _("found hosts section"));
          section = HOSTS;
        } else {
          if (section == INIT) {
            qprint("%s\n", _("Something uncommented found before OPTIONS section!"));
            fclose(cfg_file);
            exit(1);
          }
          /* Parse one line from config file */
          parse_line(line, name, value);

          if (section == HOSTS) {
            /* HOST SECTION */
            qprint("IP: %s, %s: %s\n", _("name"), name, value);
            if (valid_ip(name)) {
              qprint("%s: %s\n", _("Invalid IP"), name);
              exit(1);
            }
            for (host_idx = host_start; (host_idx < host_max) &&
                (strcmp(inet_ntoa(host_info[host_idx].addr), name) != 0) ; host_idx++);

            qprint("host_idx: %d, host_max: %d\n", host_idx, host_max);
            /* if duplicities found, skip to another record */
            if (host_idx < host_max) {
              /* qprint has condition inside. It must be braced if used in another "if"! */
              qprint("%s %s\n", name, _("is already initialized."));
            } else {
              /* IP valid -- add host */
              /* create socket */
              if ((host_info[host_max].rawfd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
                qprint("%s: %s\n%s\n", _("Initializing socket"), strerror (errno), _("Make sure you have root privileges or suid bit set!"));
                exit(1);
              } 

              qprint("socket vytvoren\n");
              if (host_info[host_max].rawfd <= 2) {
                qprint("%s\n", _("Error: got free socket less than 2!"));
                exit (1); 
              }


              /* REMEMBER, cfg_file is in NAME = VALUE format. And when IP is
               * specified, it is: XXX.XXX.XXX.XXX = www.something.com. So function
               * parse_line returns variable "name" containing IP and variable
               * "value" containing NAME of host ! */

              /* add address name into struct (value = name of the host and some parameters) */
              if (parse_value(value, val_name, &ok_delay, &noreply_delay, &long_delay)) 
                qprint("Cannot parse line\n");
                
              /* zero delay values = use default */
              if (ok_delay != 0) {
                qprint("%s: %d ms\n", _("Found ok delay"), ok_delay);
                host_info[host_max].ok_delay = ok_delay;
              } else host_info[host_max].ok_delay = delay_between_pings_ms;
              if (noreply_delay != 0) {
                qprint("%s: %d ms\n", _("Found noreply delay"), noreply_delay);
                host_info[host_max].noreply_delay = noreply_delay;
              } else host_info[host_max].noreply_delay = host_timeout_ms;
              if (val_name != NULL) qprint("%s: %s\n", _("Found name"), val_name);
              if (long_delay != 0) {
                qprint("%s: %d ms\n", _("Found long delay"), long_delay);
                host_info[host_max].long_delay = long_delay;
              } else host_info[host_max].long_delay = long_delay_ms;

              strcpy(host_info[host_max].name, val_name);

              /* add address into struct */
              if (inet_aton(name, &(host_info[host_max].addr)) == 0) {
                qprint("%s %s!\n", _("Error: Invalid host"), host_info[host_max].name);
                exit(1);
              }

              /* name = IP of the host */
              qprint("%s %s - %s\n", _("Adding host"), name, host_info[host_max].name);

              /* zero packet sequence numbers */
              host_info[host_max].last_seq_sent = 0;
              host_info[host_max].last_seq_recv = 0;
              /* zero sent/recv */
              host_info[host_max].nr_sent = 0;
              host_info[host_max].nr_recv = 0;
              /* zero last ok time */
              host_info[host_max].lastok_tv.tv_sec = 0;
              host_info[host_max].lastok_tv.tv_usec = 0;

              /* reset delay times */
              host_info[host_max].delay = 0;

              /* assign current set_nr to the host */
              host_info[host_max].set_nr = set_nr;
              host_info[host_max].id_in_set = id_in_set++;

              host_info[host_max].id = host_max;
              host_max++;
              host_ininitalized_curr_file++;
            }
          } else if (section == OPTIONS) {
            qprint("%s: %s, %s: %s\n", _("Name"), name, _("value"), value);
            /* Process options */
            if (!check_opt_value(value)) {
              int_value = atoi(value);

              /* here will options be processed */
              if (!strcmp(name, "refresh_interval_ms")) {
                refresh_interval_ms = int_value;
                qprint("%s: %s ms\n", _("Setting refresh interval to new value"), value);
              }
              if (!strcmp(name, "host_timeout_ms")) {
                host_timeout_ms = int_value;
                qprint("%s: %s ms\n", _("Setting host timeout to new value"), value);
              }
              if (!strcmp(name, "delay_between_pings_ms")) {
                delay_between_pings_ms = int_value;
                qprint("%s: %s ms\n", _("Setting delay between pings to new value"), value);
              }
              if (!strcmp(name, "long_delay_ms")) {
                long_delay_ms = int_value;
                qprint("%s: %s ms\n", _("Setting long delay to new value"), value);
              }

            } else qprint("%s '%s' \n", _("Cannot evaluate OPTIONS value of"), value);

            if (!strcmp(name, "title")) {
              qprint("%s: %s\n", _("Setting title to new value"), value);
              if (recent_title == NULL) 
                recent_title = malloc(strlen(value) + 1);
              else 
                if (strlen(recent_title) < strlen(value)) 
                  realloc(recent_title, strlen(value) + 1);
              strcpy(recent_title, value);
                  
            }
          }
        }
      }
    } /* end reading current config file */

    /* if default config file read, quit parsing */
    if (cfg_filenames == NULL) end_parsing = 1;
    if (cfgfilesidx != NULL) {
      cfgfilesidx = cfgfilesidx->next;
      if (cfgfilesidx == NULL) end_parsing = 1;
      else strncpy(cfg_path, cfgfilesidx->filename, sizeof(cfg_path) - 1);
      /* found anything in current file (set) ? */
      if (host_ininitalized_curr_file > 0) {
        set_nr++;
        id_in_set = 0;
      }
    }
    
    /* set title */
    if (host_ininitalized_curr_file > 0) {
      if (hosts->titles == NULL) {
        hosts->titles = malloc(sizeof(titlelist));
        tlist = hosts->titles;
      } else {
        tlist->next = malloc(sizeof(titlelist));
        tlist = tlist->next;
      }
      tlist->next = NULL;
      stdcntlen = 0;
      /* if no title found, set standard title */
      if (recent_title == NULL) {
        if (std_title_cnt < 10) stdcntlen = 1;
        /* we don't suppose to have more than 2 ciphers for std_title_nr */
        else stdcntlen = MIN_TITLE_LEN;
        std_title_cnt++;
        recent_title = malloc(strlen(STD_TITLE) + stdcntlen);
        strcpy(recent_title, STD_TITLE);
      }
      if (stdcntlen) {
        stdcntlen++; /* if there is a number, don't forget to have space before it */
        sprintf(tmpstr, "%d", std_title_cnt % 100);
        /* test for minimal short title like stcntlen = ".. nr" */
        if ((stdcntlen + 2) > MAX_TITLE_LEN){
          tlist->title = malloc(MAX_TITLE_LEN + 1);
          strcat(tlist->title, tmpstr);
        }
      }

      /* title is bigger than allowed and title still not set */
      if ((tlist->title == NULL) && (strlen(recent_title) > MAX_TITLE_LEN)) {
        tlist->title = malloc(MAX_TITLE_LEN + 1);
        strncpy(tlist->title, recent_title, MAX_TITLE_LEN - 2 - stdcntlen);
        tlist->title[MAX_TITLE_LEN - 3 - stdcntlen] = 0;
        strcat(tlist->title, "..");
        if (stdcntlen) strcat(tlist->title, tmpstr);
      } else {
        tlist->title = malloc(strlen(recent_title) + 1);
        strcpy(tlist->title, recent_title);
      }
      tlist->nr = tlistcnt++;
      if (tlistcnt >= MAX_TABS) {
        printf("%s\n", _("Warning: maximum allowed configuration files exceeded!"));
        end_parsing = 1;
      }
      free(recent_title);
      recent_title = NULL;
    }
  }
  

  if (host_max == MAX_HOSTS) {
    qprint("%s (%d). %s\n", _("Maximum count of hosts reached"), MAX_HOSTS, \
        _("Modify your copy of pinger if you want more hosts (requires recompile)."));
  }

  /* remember total hosts count to next parts of program */
  hosts->host_max = host_max;
  qprint("hosts:%d\n", host_max);
  
  if (cfg_file_opened) fclose(cfg_file);
  qprint("------------------\n");
}
