/*
    pinger -- GTK+/ncurses multi-ping utility
    Copyright (C) 2002 Petr Koloros

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "globals.h"
#include "ping.h"
#include "timefunc.h"
#include <netinet/ip_icmp.h>
#include <netinet/ip.h>
#include <errno.h>
#include <stdio.h>

/* Compute checksum for ICMP output buffer */
static u_short cksum (u_short *buf, u_int len)
{
	u_int sum = 0;
	u_short retval;

	/* 32 bit sum adds sequential 16 bit buffer parts */
	while (len > 1) {
		sum += *buf++;
		len -= 2;
	}
	
	/* if only one byte pending, set the half of word to zero and other half to this byte */
	if (len == 1) {

		union {
			u_short word;
			u_char byte;
		} odd;

		odd.word = 0;
		odd.byte = *(u_char *) buf;
		sum += odd.word;
	}

	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += (sum >> 16);
	retval = ~sum;
	return retval;
}


/* send ICMP echo_request to specified host */
int send_echo(host_data *host) {

	struct sockaddr_in to;
	unsigned char outpacket[sizeof (struct icmphdr)];
	struct icmphdr *ich;
	int retval;

	memset(&to, '\0', sizeof(to));
	to.sin_port = 0;
	to.sin_family = AF_INET;
	to.sin_addr = host->addr;
	ich = (struct icmphdr *) outpacket;
	ich->type = ICMP_ECHO;
	ich->code = 0;
	ich->checksum = 0;
	ich->un.echo.sequence = ++host->last_seq_sent;
	ich->un.echo.id = getpid() & 0xFFFF;
	ich->checksum = cksum ((u_short *) outpacket, sizeof (outpacket));

	set_actual_time(&(host->start));

	if ((sendto (host->rawfd, outpacket, sizeof(outpacket), 0, (struct sockaddr *)&to, sizeof(to))) < 0) {
#ifdef DEBUG
		printf("%s - %s\n", inet_ntoa(host->addr), strerror(errno));
#endif
		host->status = NW_UNREACH;
		retval = NW_UNREACH;
	} else {
    host->nr_sent++;
    retval = ECHO_OK;
  }
	return retval;
}

/* send initial ICMP echo to all host */
void pinger_init(host_data * host_info, int host_max) {
	
	int host_cnt;
	
	for (host_cnt = 0; host_cnt < host_max; host_cnt++ ) {
		if (send_echo(&(host_info[host_cnt]))) {
//			printf("%s - network is unreachable\n", inet_ntoa(host_info[host_cnt].addr));
			host_info[host_cnt].status = NW_UNREACH;
		}
		else host_info[host_cnt].status = NO_REPLY;		/* initial status = offline */
		set_actual_time(&(host_info[host_cnt].start));

	}
}


/* process icmp reply - respond to the answer and send another ICMP echo if necessary */
void process_icmp_reply(host_data *host_info) {
	
	struct sockaddr_in from;
	socklen_t fromlen = sizeof (from);
	struct iphdr *iph;
	struct icmphdr *ich;
	unsigned char inpacket[IP_MAXPACKET];
	int nrecv;
	double ping_delay;
	struct timeval stop, actual_time;


  memcpy(&stop, &saved_tv, sizeof(struct timeval));
	
	/* receive packet */
	nrecv = recvfrom (host_info->rawfd, inpacket, sizeof (inpacket), 0, (struct sockaddr *)&from, &fromlen);
	if (nrecv < 0) {
		printf("recvfrom: %s", strerror (errno));
		exit(1);
	}

	/* test for packet validity */
	if ((size_t) nrecv < sizeof (struct iphdr)) {
		fprintf(stderr, "%s\n", _("Packet too short!"));
	}

	iph = (struct iphdr *) inpacket;

	if (iph->version != 4) {
		fprintf(stderr, "%s\n", _("Wrong IP version!"));
		exit(1);
	}
	if (iph->protocol != IPPROTO_ICMP) {
		fprintf(stderr, "\n", _("Packet is no ICMP!"));
		exit(1);
	}
  if (iph->ihl * 4 + sizeof (struct icmphdr) > (size_t) nrecv) {
		fprintf(stderr, "%s\n", _("Packet too short!"));
		exit(1);
	}
	ich = (struct icmphdr *) (inpacket + iph->ihl * 4);

	/* print result */
	if (from.sin_addr.s_addr == host_info->addr.s_addr) {
		/* this was integer time meassrement
		ping_delay = tv_to_ms(&stop) - tv_to_ms(&(host_info->start));
		*/
		tvsub(&stop, &(host_info->start));
		ping_delay = (double)(stop.tv_sec) * 1000.0 +
	    				(double)(stop.tv_usec) / 1000.0;
		if (ich->un.echo.id == getpid()) {
			if (ich->type == ICMP_ECHOREPLY) {
#ifdef DEBUG
				printf("%s: OK, %.3f ms\n", inet_ntoa(host_info->addr), ping_delay);
#endif
/*				if (host_info->last_seq_sent < ich->un.echo.sequence) {
					fprintf(stderr, "%s\n", _("ICMP echo reply from the future received."));
				}
*/
				if (host_info->last_seq_sent = ich->un.echo.sequence) {
				
          if (ping_delay > host_info->long_delay)
              host_info->status = DELAYED;
          else
              host_info->status = ECHO_OK;
					host_info->delay = ping_delay;
          host_info->lastok_tv = host_info->start;
          host_info->nr_recv++;

					/* show_status(host_info); */

					/* Now the ping was OK and we want to wait delay_between_pings_ms ms.
					 * We will use offline mechanism wrote bellow to acquire the right time
					 * for sending next ping to this host after the delay.
					 * For Example: host_timeout = 3s, delay = 500ms, so we want to set
					 * host 'start' timer to (500ms - 3s) = -2,5 seconds. This causes the
					 * host to timeout after 500ms in 'offline timeout mechanism' which
					 * is set to timeout everything longer than 3 seconds. */
					set_actual_time(&actual_time);
					ms_to_tv(tv_to_ms(&actual_time) - host_info->noreply_delay + host_info->ok_delay, &(host_info->start));
				}
			} else if (ich->type == ICMP_ECHO) {
				/* show_status(host_info); */
#ifdef DEBUG
				printf("%s: LOOPBACK OK\n", inet_ntoa(host_info->addr));
#endif			
			}
		}
	}
}

/* Check all hosts for offline timeouts */
void timeout_check(hosts_data *hosts) {
	
	host_data * host_info;
	int host_max, host_cnt;
	struct timeval actual_time;
	
	host_info = hosts->host_info;
	host_max = hosts->host_max;

	/* offline timeout mechanism. Check every timer if longer than allowed amount
	 * of time. If longer, claim it offline and try send another ping. If status is 
	 * ECHO_OK, it means that host waited to send another ping. Only NO_REPLY packets can
	 * be claimed as offline
	 */
	set_actual_time(&actual_time);
	for (host_cnt = 0; host_cnt < host_max; host_cnt++) {
		if ((tv_to_ms(&actual_time) - tv_to_ms(&(host_info[host_cnt].start))) > host_info[host_cnt].noreply_delay ) {
#ifdef DEBUG
			if (host_info[host_cnt].status == NO_REPLY) 
				printf("%s: Offline\n", inet_ntoa(host_info[host_cnt].addr));
#endif
			/* Every resent packet has NO_REPLY status. If there is no answer in 
			 * host_timeout_ms ms, it means, that we still got host with NO_REPLY
			 * status and it means it.
			 */
			host_info[host_cnt].status = NO_REPLY;
			/* show_status(&(host_info[host_cnt])); */
			send_echo(&(host_info[host_cnt]));
		}
	}
}


/* Free allocated raw sockets */
int free_sockets(gpointer data)  {

	hosts_data * hosts;
	host_data * host_info;
	int host_max;
	int retval;

	retval = 0;
	
	hosts = (hosts_data *) data;
	host_info = hosts->host_info;
	host_max = hosts->host_max;

	do {
			printf("%s %s\n", _("Freeing socket for"), inet_ntoa(host_info[host_max-1].addr));
			if (close(host_info[host_max-1].rawfd)) {
					printf("%s %s\n", _("Cannot free raw socket for ip"), inet_ntoa(host_info[host_max-1].addr));
					retval = 1;
			}
			host_max--;
	} while (host_max > 0);

	return retval;
}

