/*
    pinger -- GTK+/ncurses multi-ping utility
    Copyright (C) 2002 Petr Koloros

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>

/* this defines variables from globals.h as local to this file */
#define LOCAL
#include "globals.h"
#include "ping.h"
#include "timefunc.h"
#include "parse.h"

#ifdef HAVE_LIBNCURSES
#include "interface_ncurses.h"
#endif
#ifdef HAVE_GTK_2
#include "interface_gtk.h"
#endif

#define AUTHOR_STR "Petr Koloros, <silk@sinus.cz>"

//#define DEBUG

/* show_status_function - like prototype, independent of output interface (gtk, ncurses, ..) */
int (*show_status)(host_data * host);

void print_startup_info() {
  qprint("\n%s %s, %s\n\n", _("Pinger version"), VERSION, AUTHOR_STR);
}

    

/* main of the pinger */
int main( int   argc,
          char *argv[] )
{
  /* pinger variables */
  host_data *host_info;
  hosts_data hosts;
  int host_cnt;
  int a;
  int host_info_shmid;
  int *parent_living;
  int parent_living_shmid;
  int childpid;
  filelist *flist = NULL, *flistidx = NULL;
  void (*interface_init)(int *argc, char **argv[], \
      char *title, hosts_data *hosts);
  void (*interface_done)();
  void (*gui_loop)(hosts_data *hosts, host_data *host_info);
  struct timeval cur_timeout;
  int selected;
  fd_set fdset;
  int fd_max;

#ifdef ENABLE_NLS
  bindtextdomain(PACKAGE, LOCALEDIR);
  textdomain(PACKAGE);
#endif

/* trying to set default mode as defines enable */
#ifndef HAVE_LIBNCURSES
  mode = GTK;
#else
  mode = NCURSES;
#endif
 
  /* sanity checks */
  if (MAX_CFGLINE_LEN < MAX_TITLE_LEN) {
    fprintf(stderr, _("MAX_TITLE_LEN must be less than MAX_CFGLINE_LEN!! Someone doesn't know how to make changes in the source code!\n"));
    exit(1);
  }
  if (strlen(STD_TITLE) > MAX_TITLE_LEN) {
    fprintf(stderr, _("STD_TITLE must be greater than MAX_TITLE_LEN!! Someone doesn't know how to make changes in the source code!\n"));
    exit(1);
  }
  if (MIN_TITLE_LEN > MAX_TITLE_LEN) {
    fprintf(stderr, _("MIN_TITLE_LEN must be greater than MAX_TITLE_LEN!! Someone doesn't know how to make changes in the source code!\n"));
    exit(1);
  }
  
  /* verbose output by default */
  quiet = 0;

  /* Check for parameters */
  if (argc > 1) {
    for (a = 1; a < argc; a++) {
      if (!strcmp(argv[a], "--gtk")) 
        mode = GTK;
            else if (!strcmp(argv[a], "--help")) {
        printf("\n%s %s - Copyright 2002-2005 %s, GPL'ed\n", _("Pinger - version"), VERSION, AUTHOR_STR);
        printf("\n%s\n", _("Usage:"));
        printf("    pinger [--gtk] [--quiet] [conffile1] [conffile2] ..\n\n");
        printf("%s\n", _("Parameters:"));
        printf("    --gtk   : %s\n", _("use GTK 2.x frontend (otherwise use ncurses)."));
        printf("    --quiet : %s\n\n", _("do not display messages into console."));
        exit(EXIT_FAILURE);
      } else if (!strcmp(argv[a], "--quiet"))
        quiet = 1;
      /* handle arg as file */
      else {
        if (flistidx == NULL) {        
          flistidx = malloc(sizeof(filelist));
          flist = flistidx;
        } else {
          flistidx->next = malloc(sizeof(filelist));
          flistidx = flistidx->next;
        }
        flistidx->filename = malloc(strlen(argv[a]) + 1);
        strcpy(flistidx->filename, argv[a]);
        flistidx->next = NULL;
      } /* end handle file */
          
    }
  }

  qprint(_("Available modes:"));
#ifdef HAVE_LIBNCURSES
  qprint(" ncurses");
#endif
#ifdef HAVE_GTK_2
  qprint(" GTK");
#endif
  qprint("\n%s ", _("Using mode:"));
  
  if (mode == GTK) {      
#ifndef HAVE_GTK_2
    fprintf(stderr, "%s\n", _("Error: GTK mode is not enabled, \
please recompile this program with enabling GTK mode."));
    exit(EXIT_FAILURE);
#endif
    qprint("GTK\n");
#ifdef HAVE_GTK_2    
    show_status = gtk_show_status;
    interface_init = gtk_interface_init;
    gui_loop = gtk_gui_loop;
    interface_done = gtk_interface_done;
#endif
    /* GTK has its own locale initialization. In other cases it must be done here. */
  } else setlocale(LC_ALL, "");

  if (mode == NCURSES) {
#ifndef HAVE_LIBNCURSES
    fprintf(stderr, "%s\n", _("Error: ncurses mode is not enabled, \
please recompile this program with enabling ncurses mode."));
    exit(EXIT_FAILURE);
#endif
    qprint("ncurses\n");
#ifdef HAVE_LIBNCURSES
    show_status = ncurses_show_status;
    interface_init = ncurses_interface_init;
    gui_loop = ncurses_gui_loop;
    interface_done = ncurses_interface_done;
#endif
  }
  
  /* some shared memory allocation for interprocess communication */
  if ((host_info_shmid = shmget(IPC_PRIVATE, sizeof(host_data)*MAX_HOSTS,
          IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR)) == -1) {
    fprintf(stderr, "%s\n", _("Error: cannot allocate shared memory for hosts."));
    exit(EXIT_FAILURE);
  }

  if ((parent_living_shmid = shmget(IPC_PRIVATE, sizeof(int), 
          IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR)) == -1) {
    fprintf(stderr, "%s\n", _("Error: cannot allocate shared memory for hosts."));
    shmctl(host_info_shmid, IPC_RMID, 0);
    exit(EXIT_FAILURE);
  }

  host_info = (host_data *)shmat(host_info_shmid, 0, 0);
  parent_living = (int *)shmat(parent_living_shmid, 0, 0);
  
  shmctl(host_info_shmid, IPC_RMID, 0);
  shmctl(parent_living_shmid, IPC_RMID, 0);
  
  /* this says that parent is living. If not, child will terminate itselfs */
  *parent_living = 1;
  
  hosts.host_info = host_info;
  hosts.host_max = 0;
  hosts.titles = NULL;

  read_config_file(&hosts, flist);
  
  if (hosts.host_max == 0) {
    fprintf(stderr, "%s\n", _("Error: no hosts loaded!"));
    exit(EXIT_FAILURE);
  }
  
  /* Print info about program to console */
  print_startup_info();

  if (setuid (getuid ()) < 0) {
    fprintf(stderr, "setuid: %s\n", strerror (errno));
    exit(1);
  }

  qprint("%s\n", _("Successfuly removed root privileges."));

  /* Initialization complete, now fork into two processes:
   * parent display: periodically displays informations about host
   * child engine: do pinging
   * Second writes to shared memory and first only reads data from it */

  if ((childpid = fork()) == 0) {
    /* ------- CHILD --------- */
    /* child begins here, let it be engine */
    
    FD_ZERO(&fdset);
    /* send initial ICMP echo to all hosts */
    pinger_init(host_info, hosts.host_max);

    for (host_cnt = 0; host_cnt < hosts.host_max; host_cnt++ ) 
      if (host_info[host_cnt].rawfd > fd_max) fd_max = host_info[host_cnt].rawfd;

    /* engine loop */
    while (*parent_living == 1) {
      /* refresh interval for display (default or defined in config file) */
      ms_to_tv(refresh_interval_ms, &cur_timeout);

      /* set all FDs */
      for (host_cnt = 0; host_cnt < hosts.host_max; host_cnt++ ) 
        if (!FD_ISSET(host_info[host_cnt].rawfd, &fdset))
            FD_SET(host_info[host_cnt].rawfd, &fdset);
      
      /* select, fd_max is highest used fd given acquired when parsing cfg file */
      if ((selected = select(fd_max+1, &fdset, NULL, NULL, &cur_timeout)) < 0 ) {
        fprintf(stderr, "%s %s\n", _("Cannot select:"), strerror(errno));
        exit(1);
      }
      
      /* remember reply time */
      set_actual_time(&saved_tv);
      
      if (selected)
        for (host_cnt = 0; host_cnt < hosts.host_max; host_cnt++ ) 
          if ((FD_ISSET(host_info[host_cnt].rawfd, &fdset)) /*&& (selected)*/) 
            process_icmp_reply(&host_info[host_cnt]);
      timeout_check(&hosts);
    }

    shmdt(host_info);
    shmdt(parent_living);
    exit(0);

  } else {
    /* ------- PARENT --------- */
    /* parent begins here, let it be gui, displaying results periodically */

    interface_init(&argc, &argv, hosts.titles->title, &hosts);
    
    gui_loop(&hosts, host_info);
  }

  interface_done();

  /* tell child to deallocate memory */
  *parent_living = 0;
  shmdt(parent_living);
  shmdt(host_info);
  
  return EXIT_SUCCESS;
}

